@@ -1,12 +1,39 @@
-from typing import Optional, Container, Callable, List, Dict, cast
+from __future__ import annotations
+
+from typing import Callable, Container, cast
 
-from mypy.types import (
-    Type, TypeVisitor, UnboundType, AnyType, NoneType, TypeVarId, Instance, TypeVarType,
-    CallableType, TupleType, TypedDictType, UnionType, Overloaded, ErasedType, PartialType,
-    DeletedType, TypeTranslator, UninhabitedType, TypeType, TypeOfAny, LiteralType, ProperType,
-    get_proper_type, get_proper_types, TypeAliasType, ParamSpecType, Parameters, UnpackType
-)
 from mypy.nodes import ARG_STAR, ARG_STAR2
+from mypy.types import (
+    AnyType,
+    CallableType,
+    DeletedType,
+    ErasedType,
+    Instance,
+    LiteralType,
+    NoneType,
+    Overloaded,
+    Parameters,
+    ParamSpecType,
+    PartialType,
+    ProperType,
+    TupleType,
+    Type,
+    TypeAliasType,
+    TypedDictType,
+    TypeOfAny,
+    TypeTranslator,
+    TypeType,
+    TypeVarId,
+    TypeVarTupleType,
+    TypeVarType,
+    TypeVisitor,
+    UnboundType,
+    UninhabitedType,
+    UnionType,
+    UnpackType,
+    get_proper_type,
+    get_proper_types,
+)
 
 
 def erase_type(typ: Type) -> ProperType:
@@ -26,7 +53,6 @@ def erase_type(typ: Type) -> ProperType:
 
 
 class EraseTypeVisitor(TypeVisitor[ProperType]):
-
     def visit_unbound_type(self, t: UnboundType) -> ProperType:
         # TODO: replace with an assert after UnboundType can't leak from semantic analysis.
         return AnyType(TypeOfAny.from_error)
@@ -62,8 +88,11 @@ class EraseTypeVisitor(TypeVisitor[ProperType]):
     def visit_parameters(self, t: Parameters) -> ProperType:
         raise RuntimeError("Parameters should have been bound to a class")
 
+    def visit_type_var_tuple(self, t: TypeVarTupleType) -> ProperType:
+        return AnyType(TypeOfAny.special_form)
+
     def visit_unpack_type(self, t: UnpackType) -> ProperType:
-        raise NotImplementedError
+        return AnyType(TypeOfAny.special_form)
 
     def visit_callable_type(self, t: CallableType) -> ProperType:
         # We must preserve the fallback type for overload resolution to work.
@@ -96,6 +125,7 @@ class EraseTypeVisitor(TypeVisitor[ProperType]):
     def visit_union_type(self, t: UnionType) -> ProperType:
         erased_items = [erase_type(item) for item in t.items]
         from mypy.typeops import make_simplified_union
+
         return make_simplified_union(erased_items)
 
     def visit_type_type(self, t: TypeType) -> ProperType:
@@ -105,14 +135,16 @@ class EraseTypeVisitor(TypeVisitor[ProperType]):
         raise RuntimeError("Type aliases should be expanded before accepting this visitor")
 
 
-def erase_typevars(t: Type, ids_to_erase: Optional[Container[TypeVarId]] = None) -> Type:
+def erase_typevars(t: Type, ids_to_erase: Container[TypeVarId] | None = None) -> Type:
     """Replace all type variables in a type with any,
     or just the ones in the provided collection.
     """
+
     def erase_id(id: TypeVarId) -> bool:
         if ids_to_erase is None:
             return True
         return id in ids_to_erase
+
     return t.accept(TypeVarEraser(erase_id, AnyType(TypeOfAny.special_form)))
 
 
@@ -133,6 +165,11 @@ class TypeVarEraser(TypeTranslator):
             return self.replacement
         return t
 
+    def visit_type_var_tuple(self, t: TypeVarTupleType) -> Type:
+        if self.erase_id(t.id):
+            return self.replacement
+        return t
+
     def visit_param_spec(self, t: ParamSpecType) -> Type:
         if self.erase_id(t.id):
             return self.replacement
@@ -155,13 +192,7 @@ class LastKnownValueEraser(TypeTranslator):
     def visit_instance(self, t: Instance) -> Type:
         if not t.last_known_value and not t.args:
             return t
-        new_t = t.copy_modified(
-            args=[a.accept(self) for a in t.args],
-            last_known_value=None,
-        )
-        new_t.can_be_true = t.can_be_true
-        new_t.can_be_false = t.can_be_false
-        return new_t
+        return t.copy_modified(args=[a.accept(self) for a in t.args], last_known_value=None)
 
     def visit_type_alias_type(self, t: TypeAliasType) -> Type:
         # Type aliases can't contain literal values, because they are
@@ -174,17 +205,18 @@ class LastKnownValueEraser(TypeTranslator):
         # Call make_simplified_union only on lists of instance types
         # that all have the same fullname, to avoid simplifying too
         # much.
-        instances = [item for item in new.items
-                     if isinstance(get_proper_type(item), Instance)]
+        instances = [item for item in new.items if isinstance(get_proper_type(item), Instance)]
         # Avoid merge in simple cases such as optional types.
         if len(instances) > 1:
-            instances_by_name: Dict[str, List[Instance]] = {}
-            new_items = get_proper_types(new.items)
-            for item in new_items:
-                if isinstance(item, Instance) and not item.args:
-                    instances_by_name.setdefault(item.type.fullname, []).append(item)
-            merged: List[Type] = []
-            for item in new_items:
+            instances_by_name: dict[str, list[Instance]] = {}
+            p_new_items = get_proper_types(new.items)
+            for p_item in p_new_items:
+                if isinstance(p_item, Instance) and not p_item.args:
+                    instances_by_name.setdefault(p_item.type.fullname, []).append(p_item)
+            merged: list[Type] = []
+            for item in new.items:
+                orig_item = item
+                item = get_proper_type(item)
                 if isinstance(item, Instance) and not item.args:
                     types = instances_by_name.get(item.type.fullname)
                     if types is not None:
@@ -192,9 +224,10 @@ class LastKnownValueEraser(TypeTranslator):
                             merged.append(item)
                         else:
                             from mypy.typeops import make_simplified_union
+
                             merged.append(make_simplified_union(types))
                             del instances_by_name[item.type.fullname]
                 else:
-                    merged.append(item)
+                    merged.append(orig_item)
             return UnionType.make_union(merged)
         return new
