@@ -1,35 +1,60 @@
 """Plugin that provides support for dataclasses."""
 
-from typing import Dict, List, Set, Tuple, Optional
+from __future__ import annotations
+
 from typing_extensions import Final
 
 from mypy.nodes import (
-    ARG_OPT, ARG_NAMED, ARG_NAMED_OPT, ARG_POS, ARG_STAR, ARG_STAR2, MDEF,
-    Argument, AssignmentStmt, CallExpr,    Context, Expression, JsonDict,
-    NameExpr, RefExpr, SymbolTableNode, TempNode, TypeInfo, Var, TypeVarExpr,
-    PlaceholderNode
+    ARG_NAMED,
+    ARG_NAMED_OPT,
+    ARG_OPT,
+    ARG_POS,
+    ARG_STAR,
+    ARG_STAR2,
+    MDEF,
+    Argument,
+    AssignmentStmt,
+    CallExpr,
+    Context,
+    Expression,
+    JsonDict,
+    NameExpr,
+    PlaceholderNode,
+    RefExpr,
+    SymbolTableNode,
+    TempNode,
+    TypeAlias,
+    TypeInfo,
+    TypeVarExpr,
+    Var,
 )
 from mypy.plugin import ClassDefContext, SemanticAnalyzerPluginInterface
 from mypy.plugins.common import (
-    add_method, _get_decorator_bool_argument, deserialize_and_fixup_type, add_attribute_to_class,
-)
-from mypy.typeops import map_type_from_supertype
-from mypy.types import (
-    Type, Instance, NoneType, TypeVarType, CallableType, TupleType, LiteralType,
-    get_proper_type, AnyType, TypeOfAny,
+    _get_decorator_bool_argument,
+    add_attribute_to_class,
+    add_method,
+    deserialize_and_fixup_type,
 )
 from mypy.server.trigger import make_wildcard_trigger
 from mypy.state import state
+from mypy.typeops import map_type_from_supertype
+from mypy.types import (
+    AnyType,
+    CallableType,
+    Instance,
+    LiteralType,
+    NoneType,
+    TupleType,
+    Type,
+    TypeOfAny,
+    TypeVarType,
+    get_proper_type,
+)
 
 # The set of decorators that generate dataclasses.
-dataclass_makers: Final = {
-    'dataclass',
-    'dataclasses.dataclass',
-}
+dataclass_makers: Final = {"dataclass", "dataclasses.dataclass"}
 # The set of functions that generate dataclass fields.
-field_makers: Final = {
-    'dataclasses.field',
-}
+field_makers: Final = {"dataclasses.field"}
 
 
 SELF_TVAR_NAME: Final = "_DT"
@@ -37,16 +62,16 @@ SELF_TVAR_NAME: Final = "_DT"
 
 class DataclassAttribute:
     def __init__(
-            self,
-            name: str,
-            is_in_init: bool,
-            is_init_var: bool,
-            has_default: bool,
-            line: int,
-            column: int,
-            type: Optional[Type],
-            info: TypeInfo,
-            kw_only: bool,
+        self,
+        name: str,
+        is_in_init: bool,
+        is_init_var: bool,
+        has_default: bool,
+        line: int,
+        column: int,
+        type: Type | None,
+        info: TypeInfo,
+        kw_only: bool,
     ) -> None:
         self.name = name
         self.is_in_init = is_in_init
@@ -67,10 +92,7 @@ class DataclassAttribute:
         elif not self.kw_only and self.has_default:
             arg_kind = ARG_OPT
         return Argument(
-            variable=self.to_var(),
-            type_annotation=self.type,
-            initializer=None,
-            kind=arg_kind,
+            variable=self.to_var(), type_annotation=self.type, initializer=None, kind=arg_kind
         )
 
     def to_var(self) -> Var:
@@ -79,24 +101,24 @@ class DataclassAttribute:
     def serialize(self) -> JsonDict:
         assert self.type
         return {
-            'name': self.name,
-            'is_in_init': self.is_in_init,
-            'is_init_var': self.is_init_var,
-            'has_default': self.has_default,
-            'line': self.line,
-            'column': self.column,
-            'type': self.type.serialize(),
-            'kw_only': self.kw_only,
+            "name": self.name,
+            "is_in_init": self.is_in_init,
+            "is_init_var": self.is_init_var,
+            "has_default": self.has_default,
+            "line": self.line,
+            "column": self.column,
+            "type": self.type.serialize(),
+            "kw_only": self.kw_only,
         }
 
     @classmethod
     def deserialize(
         cls, info: TypeInfo, data: JsonDict, api: SemanticAnalyzerPluginInterface
-    ) -> 'DataclassAttribute':
+    ) -> DataclassAttribute:
         data = data.copy()
-        if data.get('kw_only') is None:
-            data['kw_only'] = False
-        typ = deserialize_and_fixup_type(data.pop('type'), api)
+        if data.get("kw_only") is None:
+            data["kw_only"] = False
+        typ = deserialize_and_fixup_type(data.pop("type"), api)
         return cls(type=typ, info=info, **data)
 
     def expand_typevar_from_subtype(self, sub_type: TypeInfo) -> None:
@@ -107,10 +129,19 @@ class DataclassAttribute:
 
 
 class DataclassTransformer:
+    """Implement the behavior of @dataclass.
+
+    Note that this may be executed multiple times on the same class, so
+    everything here must be idempotent.
+
+    This runs after the main semantic analysis pass, so you can assume that
+    there are no placeholders.
+    """
+
     def __init__(self, ctx: ClassDefContext) -> None:
         self._ctx = ctx
 
-    def transform(self) -> None:
+    def transform(self) -> bool:
         """Apply all the necessary transformations to the underlying
         dataclass so as to ensure it is fully type checked according
         to the rules in PEP 557.
@@ -119,19 +150,18 @@ class DataclassTransformer:
         info = self._ctx.cls.info
         attributes = self.collect_attributes()
         if attributes is None:
-            # Some definitions are not ready, defer() should be already called.
-            return
+            # Some definitions are not ready. We need another pass.
+            return False
         for attr in attributes:
             if attr.type is None:
-                ctx.api.defer()
-                return
+                return False
         decorator_arguments = {
-            'init': _get_decorator_bool_argument(self._ctx, 'init', True),
-            'eq': _get_decorator_bool_argument(self._ctx, 'eq', True),
-            'order': _get_decorator_bool_argument(self._ctx, 'order', False),
-            'frozen': _get_decorator_bool_argument(self._ctx, 'frozen', False),
-            'slots': _get_decorator_bool_argument(self._ctx, 'slots', False),
-            'match_args': _get_decorator_bool_argument(self._ctx, 'match_args', True),
+            "init": _get_decorator_bool_argument(self._ctx, "init", True),
+            "eq": _get_decorator_bool_argument(self._ctx, "eq", True),
+            "order": _get_decorator_bool_argument(self._ctx, "order", False),
+            "frozen": _get_decorator_bool_argument(self._ctx, "frozen", False),
+            "slots": _get_decorator_bool_argument(self._ctx, "slots", False),
+            "match_args": _get_decorator_bool_argument(self._ctx, "match_args", True),
         }
         py_version = self._ctx.api.options.python_version
 
@@ -139,12 +169,17 @@ class DataclassTransformer:
         # processed them yet. In order to work around this, we can simply skip generating
         # __init__ if there are no attributes, because if the user truly did not define any,
         # then the object default __init__ with an empty signature will be present anyway.
-        if (decorator_arguments['init'] and
-                ('__init__' not in info.names or info.names['__init__'].plugin_generated) and
-                attributes):
+        if (
+            decorator_arguments["init"]
+            and ("__init__" not in info.names or info.names["__init__"].plugin_generated)
+            and attributes
+        ):
 
-            args = [attr.to_argument() for attr in attributes if attr.is_in_init
-                    and not self._is_kw_only_type(attr.type)]
+            args = [
+                attr.to_argument()
+                for attr in attributes
+                if attr.is_in_init and not self._is_kw_only_type(attr.type)
+            ]
 
             if info.fallback_to_any:
                 # Make positional args optional since we don't know their order.
@@ -154,48 +189,49 @@ class DataclassTransformer:
                     if arg.kind == ARG_POS:
                         arg.kind = ARG_OPT
 
-                nameless_var = Var('')
-                args = [Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR),
-                        *args,
-                        Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR2),
-                        ]
+                nameless_var = Var("")
+                args = [
+                    Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR),
+                    *args,
+                    Argument(nameless_var, AnyType(TypeOfAny.explicit), None, ARG_STAR2),
+                ]
 
-            add_method(
-                ctx,
-                '__init__',
-                args=args,
-                return_type=NoneType(),
-            )
+            add_method(ctx, "__init__", args=args, return_type=NoneType())
 
-        if (decorator_arguments['eq'] and info.get('__eq__') is None or
-                decorator_arguments['order']):
+        if (
+            decorator_arguments["eq"]
+            and info.get("__eq__") is None
+            or decorator_arguments["order"]
+        ):
             # Type variable for self types in generated methods.
-            obj_type = ctx.api.named_type('builtins.object')
-            self_tvar_expr = TypeVarExpr(SELF_TVAR_NAME, info.fullname + '.' + SELF_TVAR_NAME,
-                                         [], obj_type)
+            obj_type = ctx.api.named_type("builtins.object")
+            self_tvar_expr = TypeVarExpr(
+                SELF_TVAR_NAME, info.fullname + "." + SELF_TVAR_NAME, [], obj_type
+            )
             info.names[SELF_TVAR_NAME] = SymbolTableNode(MDEF, self_tvar_expr)
 
         # Add <, >, <=, >=, but only if the class has an eq method.
-        if decorator_arguments['order']:
-            if not decorator_arguments['eq']:
-                ctx.api.fail('eq must be True if order is True', ctx.cls)
+        if decorator_arguments["order"]:
+            if not decorator_arguments["eq"]:
+                ctx.api.fail("eq must be True if order is True", ctx.cls)
 
-            for method_name in ['__lt__', '__gt__', '__le__', '__ge__']:
+            for method_name in ["__lt__", "__gt__", "__le__", "__ge__"]:
                 # Like for __eq__ and __ne__, we want "other" to match
                 # the self type.
-                obj_type = ctx.api.named_type('builtins.object')
-                order_tvar_def = TypeVarType(SELF_TVAR_NAME, info.fullname + '.' + SELF_TVAR_NAME,
-                                            -1, [], obj_type)
-                order_return_type = ctx.api.named_type('builtins.bool')
+                obj_type = ctx.api.named_type("builtins.object")
+                order_tvar_def = TypeVarType(
+                    SELF_TVAR_NAME, info.fullname + "." + SELF_TVAR_NAME, -1, [], obj_type
+                )
+                order_return_type = ctx.api.named_type("builtins.bool")
                 order_args = [
-                    Argument(Var('other', order_tvar_def), order_tvar_def, None, ARG_POS)
+                    Argument(Var("other", order_tvar_def), order_tvar_def, None, ARG_POS)
                 ]
 
                 existing_method = info.get(method_name)
                 if existing_method is not None and not existing_method.plugin_generated:
                     assert existing_method.node
                     ctx.api.fail(
-                        'You may not have a custom %s method when order=True' % method_name,
+                        f"You may not have a custom {method_name} method when order=True",
                         existing_method.node,
                     )
 
@@ -208,59 +244,75 @@ class DataclassTransformer:
                     tvar_def=order_tvar_def,
                 )
 
-        if decorator_arguments['frozen']:
+        parent_decorator_arguments = []
+        for parent in info.mro[1:-1]:
+            parent_args = parent.metadata.get("dataclass")
+            if parent_args:
+                parent_decorator_arguments.append(parent_args)
+
+        if decorator_arguments["frozen"]:
+            if any(not parent["frozen"] for parent in parent_decorator_arguments):
+                ctx.api.fail("Cannot inherit frozen dataclass from a non-frozen one", info)
             self._propertize_callables(attributes, settable=False)
             self._freeze(attributes)
         else:
+            if any(parent["frozen"] for parent in parent_decorator_arguments):
+                ctx.api.fail("Cannot inherit non-frozen dataclass from a frozen one", info)
             self._propertize_callables(attributes)
 
-        if decorator_arguments['slots']:
+        if decorator_arguments["slots"]:
             self.add_slots(info, attributes, correct_version=py_version >= (3, 10))
 
         self.reset_init_only_vars(info, attributes)
 
-        if (decorator_arguments['match_args'] and
-                ('__match_args__' not in info.names or
-                 info.names['__match_args__'].plugin_generated) and
-                attributes):
+        if (
+            decorator_arguments["match_args"]
+            and (
+                "__match_args__" not in info.names or info.names["__match_args__"].plugin_generated
+            )
+            and attributes
+            and py_version >= (3, 10)
+        ):
             str_type = ctx.api.named_type("builtins.str")
-            literals: List[Type] = [LiteralType(attr.name, str_type)
-                        for attr in attributes if attr.is_in_init]
+            literals: list[Type] = [
+                LiteralType(attr.name, str_type) for attr in attributes if attr.is_in_init
+            ]
             match_args_type = TupleType(literals, ctx.api.named_type("builtins.tuple"))
             add_attribute_to_class(ctx.api, ctx.cls, "__match_args__", match_args_type)
 
         self._add_dataclass_fields_magic_attribute()
 
-        info.metadata['dataclass'] = {
-            'attributes': [attr.serialize() for attr in attributes],
-            'frozen': decorator_arguments['frozen'],
+        info.metadata["dataclass"] = {
+            "attributes": [attr.serialize() for attr in attributes],
+            "frozen": decorator_arguments["frozen"],
         }
 
-    def add_slots(self,
-                  info: TypeInfo,
-                  attributes: List[DataclassAttribute],
-                  *,
-                  correct_version: bool) -> None:
+        return True
+
+    def add_slots(
+        self, info: TypeInfo, attributes: list[DataclassAttribute], *, correct_version: bool
+    ) -> None:
         if not correct_version:
             # This means that version is lower than `3.10`,
             # it is just a non-existent argument for `dataclass` function.
             self._ctx.api.fail(
                 'Keyword argument "slots" for "dataclass" '
-                'is only valid in Python 3.10 and higher',
+                "is only valid in Python 3.10 and higher",
                 self._ctx.reason,
             )
             return
 
         generated_slots = {attr.name for attr in attributes}
-        if ((info.slots is not None and info.slots != generated_slots)
-                or info.names.get('__slots__')):
+        if (info.slots is not None and info.slots != generated_slots) or info.names.get(
+            "__slots__"
+        ):
             # This means we have a slots conflict.
             # Class explicitly specifies a different `__slots__` field.
             # And `@dataclass(slots=True)` is used.
             # In runtime this raises a type error.
             self._ctx.api.fail(
                 '"{}" both defines "__slots__" and is used with "slots=True"'.format(
-                    self._ctx.cls.name,
+                    self._ctx.cls.name
                 ),
                 self._ctx.cls,
             )
@@ -268,7 +320,7 @@ class DataclassTransformer:
 
         info.slots = generated_slots
 
-    def reset_init_only_vars(self, info: TypeInfo, attributes: List[DataclassAttribute]) -> None:
+    def reset_init_only_vars(self, info: TypeInfo, attributes: list[DataclassAttribute]) -> None:
         """Remove init-only vars from the class and reset init var declarations."""
         for attr in attributes:
             if attr.is_init_var:
@@ -285,7 +337,7 @@ class DataclassTransformer:
                             # recreate a symbol node for this attribute.
                             lvalue.node = None
 
-    def collect_attributes(self) -> Optional[List[DataclassAttribute]]:
+    def collect_attributes(self) -> list[DataclassAttribute] | None:
         """Collect all attributes declared in the dataclass and its parents.
 
         All assignments of the form
@@ -294,13 +346,56 @@ class DataclassTransformer:
           b: SomeOtherType = ...
 
         are collected.
+
+        Return None if some dataclass base class hasn't been processed
+        yet and thus we'll need to ask for another pass.
         """
-        # First, collect attributes belonging to the current class.
         ctx = self._ctx
         cls = self._ctx.cls
-        attrs: List[DataclassAttribute] = []
-        known_attrs: Set[str] = set()
-        kw_only = _get_decorator_bool_argument(ctx, 'kw_only', False)
+
+        # First, collect attributes belonging to any class in the MRO, ignoring duplicates.
+        #
+        # We iterate through the MRO in reverse because attrs defined in the parent must appear
+        # earlier in the attributes list than attrs defined in the child. See:
+        # https://docs.python.org/3/library/dataclasses.html#inheritance
+        #
+        # However, we also want attributes defined in the subtype to override ones defined
+        # in the parent. We can implement this via a dict without disrupting the attr order
+        # because dicts preserve insertion order in Python 3.7+.
+        found_attrs: dict[str, DataclassAttribute] = {}
+        found_dataclass_supertype = False
+        for info in reversed(cls.info.mro[1:-1]):
+            if "dataclass_tag" in info.metadata and "dataclass" not in info.metadata:
+                # We haven't processed the base class yet. Need another pass.
+                return None
+            if "dataclass" not in info.metadata:
+                continue
+
+            # Each class depends on the set of attributes in its dataclass ancestors.
+            ctx.api.add_plugin_dependency(make_wildcard_trigger(info.fullname))
+            found_dataclass_supertype = True
+
+            for data in info.metadata["dataclass"]["attributes"]:
+                name: str = data["name"]
+
+                attr = DataclassAttribute.deserialize(info, data, ctx.api)
+                # TODO: We shouldn't be performing type operations during the main
+                #       semantic analysis pass, since some TypeInfo attributes might
+                #       still be in flux. This should be performed in a later phase.
+                with state.strict_optional_set(ctx.api.options.strict_optional):
+                    attr.expand_typevar_from_subtype(ctx.cls.info)
+                found_attrs[name] = attr
+
+                sym_node = cls.info.names.get(name)
+                if sym_node and sym_node.node and not isinstance(sym_node.node, Var):
+                    ctx.api.fail(
+                        "Dataclass attribute may only be overridden by another attribute",
+                        sym_node.node,
+                    )
+
+        # Second, collect attributes belonging to the current class.
+        current_attr_names: set[str] = set()
+        kw_only = _get_decorator_bool_argument(ctx, "kw_only", False)
         for stmt in cls.defs.body:
             # Any assignment that doesn't use the new type declaration
             # syntax can be ignored out of hand.
@@ -315,14 +410,22 @@ class DataclassTransformer:
 
             sym = cls.info.names.get(lhs.name)
             if sym is None:
-                # This name is likely blocked by a star import. We don't need to defer because
-                # defer() is already called by mark_incomplete().
+                # There was probably a semantic analysis error.
                 continue
 
             node = sym.node
-            if isinstance(node, PlaceholderNode):
-                # This node is not ready yet.
-                return None
+            assert not isinstance(node, PlaceholderNode)
+
+            if isinstance(node, TypeAlias):
+                ctx.api.fail(
+                    ("Type aliases inside dataclass definitions are not supported at runtime"),
+                    node,
+                )
+                # Skip processing this node. This doesn't match the runtime behaviour,
+                # but the only alternative would be to modify the SymbolTable,
+                # and it's a little hairy to do that in a plugin.
+                continue
+
             assert isinstance(node, Var)
 
             # x: ClassVar[int] is ignored by dataclasses.
@@ -332,8 +435,10 @@ class DataclassTransformer:
             # x: InitVar[int] is turned into x: int and is removed from the class.
             is_init_var = False
             node_type = get_proper_type(node.type)
-            if (isinstance(node_type, Instance) and
-                    node_type.type.fullname == 'dataclasses.InitVar'):
+            if (
+                isinstance(node_type, Instance)
+                and node_type.type.fullname == "dataclasses.InitVar"
+            ):
                 is_init_var = True
                 node.type = node_type.args[0]
 
@@ -342,7 +447,7 @@ class DataclassTransformer:
 
             has_field_call, field_args = _collect_field_args(stmt.rvalue, ctx)
 
-            is_in_init_param = field_args.get('init')
+            is_in_init_param = field_args.get("init")
             if is_in_init_param is None:
                 is_in_init = True
             else:
@@ -352,7 +457,7 @@ class DataclassTransformer:
             # Ensure that something like x: int = field() is rejected
             # after an attribute with a default.
             if has_field_call:
-                has_default = 'default' in field_args or 'default_factory' in field_args
+                has_default = "default" in field_args or "default_factory" in field_args
 
             # All other assignments are already type checked.
             elif not isinstance(stmt.rvalue, TempNode):
@@ -366,12 +471,28 @@ class DataclassTransformer:
             is_kw_only = kw_only
             # Use the kw_only field arg if it is provided. Otherwise use the
             # kw_only value from the decorator parameter.
-            field_kw_only_param = field_args.get('kw_only')
+            field_kw_only_param = field_args.get("kw_only")
             if field_kw_only_param is not None:
                 is_kw_only = bool(ctx.api.parse_bool(field_kw_only_param))
 
-            known_attrs.add(lhs.name)
-            attrs.append(DataclassAttribute(
+            if sym.type is None and node.is_final and node.is_inferred:
+                # This is a special case, assignment like x: Final = 42 is classified
+                # annotated above, but mypy strips the `Final` turning it into x = 42.
+                # We do not support inferred types in dataclasses, so we can try inferring
+                # type for simple literals, and otherwise require an explicit type
+                # argument for Final[...].
+                typ = ctx.api.analyze_simple_literal_type(stmt.rvalue, is_final=True)
+                if typ:
+                    node.type = typ
+                else:
+                    ctx.api.fail(
+                        "Need type argument for Final[...] with non-literal default in dataclass",
+                        stmt,
+                    )
+                    node.type = AnyType(TypeOfAny.from_error)
+
+            current_attr_names.add(lhs.name)
+            found_attrs[lhs.name] = DataclassAttribute(
                 name=lhs.name,
                 is_in_init=is_in_init,
                 is_init_var=is_init_var,
@@ -381,50 +502,16 @@ class DataclassTransformer:
                 type=sym.type,
                 info=cls.info,
                 kw_only=is_kw_only,
-            ))
+            )
 
-        # Next, collect attributes belonging to any class in the MRO
-        # as long as those attributes weren't already collected.  This
-        # makes it possible to overwrite attributes in subclasses.
-        # copy() because we potentially modify all_attrs below and if this code requires debugging
-        # we'll have unmodified attrs laying around.
-        all_attrs = attrs.copy()
-        for info in cls.info.mro[1:-1]:
-            if 'dataclass' not in info.metadata:
-                continue
-
-            super_attrs = []
-            # Each class depends on the set of attributes in its dataclass ancestors.
-            ctx.api.add_plugin_dependency(make_wildcard_trigger(info.fullname))
-
-            for data in info.metadata["dataclass"]["attributes"]:
-                name: str = data["name"]
-                if name not in known_attrs:
-                    attr = DataclassAttribute.deserialize(info, data, ctx.api)
-                    # TODO: We shouldn't be performing type operations during the main
-                    #       semantic analysis pass, since some TypeInfo attributes might
-                    #       still be in flux. This should be performed in a later phase.
-                    with state.strict_optional_set(ctx.api.options.strict_optional):
-                        attr.expand_typevar_from_subtype(ctx.cls.info)
-                    known_attrs.add(name)
-                    super_attrs.append(attr)
-                elif all_attrs:
-                    # How early in the attribute list an attribute appears is determined by the
-                    # reverse MRO, not simply MRO.
-                    # See https://docs.python.org/3/library/dataclasses.html#inheritance for
-                    # details.
-                    for attr in all_attrs:
-                        if attr.name == name:
-                            all_attrs.remove(attr)
-                            super_attrs.append(attr)
-                            break
-            all_attrs = super_attrs + all_attrs
+        all_attrs = list(found_attrs.values())
+        if found_dataclass_supertype:
             all_attrs.sort(key=lambda a: a.kw_only)
 
-        # Ensure that arguments without a default don't follow
-        # arguments that have a default.
+        # Third, ensure that arguments without a default don't follow
+        # arguments that have a default and that the KW_ONLY sentinel
+        # is only provided once.
         found_default = False
-        # Ensure that the KW_ONLY sentinel is only provided once
         found_kw_sentinel = False
         for attr in all_attrs:
             # If we find any attribute that is_in_init, not kw_only, and that
@@ -433,26 +520,23 @@ class DataclassTransformer:
             if found_default and attr.is_in_init and not attr.has_default and not attr.kw_only:
                 # If the issue comes from merging different classes, report it
                 # at the class definition point.
-                context = (Context(line=attr.line, column=attr.column) if attr in attrs
-                           else ctx.cls)
+                context: Context = ctx.cls
+                if attr.name in current_attr_names:
+                    context = Context(line=attr.line, column=attr.column)
                 ctx.api.fail(
-                    'Attributes without a default cannot follow attributes with one',
-                    context,
+                    "Attributes without a default cannot follow attributes with one", context
                 )
 
             found_default = found_default or (attr.has_default and attr.is_in_init)
             if found_kw_sentinel and self._is_kw_only_type(attr.type):
-                context = (Context(line=attr.line, column=attr.column) if attr in attrs
-                           else ctx.cls)
-                ctx.api.fail(
-                    'There may not be more than one field with the KW_ONLY type',
-                    context,
-                )
+                context = ctx.cls
+                if attr.name in current_attr_names:
+                    context = Context(line=attr.line, column=attr.column)
+                ctx.api.fail("There may not be more than one field with the KW_ONLY type", context)
             found_kw_sentinel = found_kw_sentinel or self._is_kw_only_type(attr.type)
-
         return all_attrs
 
-    def _freeze(self, attributes: List[DataclassAttribute]) -> None:
+    def _freeze(self, attributes: list[DataclassAttribute]) -> None:
         """Converts all attributes to @property methods in order to
         emulate frozen classes.
         """
@@ -461,18 +545,18 @@ class DataclassTransformer:
             sym_node = info.names.get(attr.name)
             if sym_node is not None:
                 var = sym_node.node
-                assert isinstance(var, Var)
-                var.is_property = True
+                if isinstance(var, Var):
+                    var.is_property = True
             else:
                 var = attr.to_var()
                 var.info = info
                 var.is_property = True
-                var._fullname = info.fullname + '.' + var.name
+                var._fullname = info.fullname + "." + var.name
                 info.names[var.name] = SymbolTableNode(MDEF, var)
 
-    def _propertize_callables(self,
-                              attributes: List[DataclassAttribute],
-                              settable: bool = True) -> None:
+    def _propertize_callables(
+        self, attributes: list[DataclassAttribute], settable: bool = True
+    ) -> None:
         """Converts all attributes with callable types to @property methods.
 
         This avoids the typechecker getting confused and thinking that
@@ -487,53 +571,61 @@ class DataclassTransformer:
                 var.info = info
                 var.is_property = True
                 var.is_settable_property = settable
-                var._fullname = info.fullname + '.' + var.name
+                var._fullname = info.fullname + "." + var.name
                 info.names[var.name] = SymbolTableNode(MDEF, var)
 
-    def _is_kw_only_type(self, node: Optional[Type]) -> bool:
+    def _is_kw_only_type(self, node: Type | None) -> bool:
         """Checks if the type of the node is the KW_ONLY sentinel value."""
         if node is None:
             return False
         node_type = get_proper_type(node)
         if not isinstance(node_type, Instance):
             return False
-        return node_type.type.fullname == 'dataclasses.KW_ONLY'
+        return node_type.type.fullname == "dataclasses.KW_ONLY"
 
     def _add_dataclass_fields_magic_attribute(self) -> None:
-        attr_name = '__dataclass_fields__'
+        attr_name = "__dataclass_fields__"
         any_type = AnyType(TypeOfAny.explicit)
-        field_type = self._ctx.api.named_type_or_none('dataclasses.Field', [any_type]) or any_type
-        attr_type = self._ctx.api.named_type('builtins.dict', [
-            self._ctx.api.named_type('builtins.str'),
-            field_type,
-        ])
+        field_type = self._ctx.api.named_type_or_none("dataclasses.Field", [any_type]) or any_type
+        attr_type = self._ctx.api.named_type(
+            "builtins.dict", [self._ctx.api.named_type("builtins.str"), field_type]
+        )
         var = Var(name=attr_name, type=attr_type)
         var.info = self._ctx.cls.info
-        var._fullname = self._ctx.cls.info.fullname + '.' + attr_name
+        var._fullname = self._ctx.cls.info.fullname + "." + attr_name
+        var.is_classvar = True
         self._ctx.cls.info.names[attr_name] = SymbolTableNode(
-            kind=MDEF,
-            node=var,
-            plugin_generated=True,
+            kind=MDEF, node=var, plugin_generated=True
         )
 
 
-def dataclass_class_maker_callback(ctx: ClassDefContext) -> None:
-    """Hooks into the class typechecking process to add support for dataclasses.
+def dataclass_tag_callback(ctx: ClassDefContext) -> None:
+    """Record that we have a dataclass in the main semantic analysis pass.
+
+    The later pass implemented by DataclassTransformer will use this
+    to detect dataclasses in base classes.
     """
+    # The value is ignored, only the existence matters.
+    ctx.cls.info.metadata["dataclass_tag"] = {}
+
+
+def dataclass_class_maker_callback(ctx: ClassDefContext) -> bool:
+    """Hooks into the class typechecking process to add support for dataclasses."""
     transformer = DataclassTransformer(ctx)
-    transformer.transform()
+    return transformer.transform()
 
 
-def _collect_field_args(expr: Expression,
-                        ctx: ClassDefContext) -> Tuple[bool, Dict[str, Expression]]:
+def _collect_field_args(
+    expr: Expression, ctx: ClassDefContext
+) -> tuple[bool, dict[str, Expression]]:
     """Returns a tuple where the first value represents whether or not
     the expression is a call to dataclass.field and the second is a
     dictionary of the keyword arguments that field() was called with.
     """
     if (
-            isinstance(expr, CallExpr) and
-            isinstance(expr.callee, RefExpr) and
-            expr.callee.fullname in field_makers
+        isinstance(expr, CallExpr)
+        and isinstance(expr.callee, RefExpr)
+        and expr.callee.fullname in field_makers
     ):
         # field() only takes keyword arguments.
         args = {}
